#!/usr/bin/env python3
''' Author: github.com/loniceraLeo
'''

import asyncio
import ssl


from TlsProxy import config, flags
from TlsProxy.config import CLIENT_SIDE
from TlsProxy.utils import *

def nop(*args, **kwargs):
    pass

async def process_stream(rd: asyncio.StreamReader, 
        wt: asyncio.StreamWriter):
    data = await rd.read(4096)
    raw_addr = data.split(b' ')[1]
    padded_addr = padding(raw_addr, conf['padding-length'])
    addr, _ = mask(padded_addr, new_key)

    try:
        rmt_rd, rmt_wt = await asyncio.open_connection(
            host=conf['server'], port=conf['port'],
            ssl=ctx, ssl_handshake_timeout=30)
    except:
        wt.close()
        return

    rmt_wt.write(addr)
    await rmt_wt.drain()

    wt.write(b'HTTP/1.1 200 Connection Established\r\n\r\n')
    await wt.drain()

    await asyncio.gather(stream_copy(rd, rmt_wt), 
                            stream_copy(rmt_rd, wt),
                            return_exceptions=True)

async def stream_copy(reader: asyncio.StreamReader, 
    writer: asyncio.StreamWriter, toggle=False):
    first = 0
    try:
        while True:
            data, first = mask(await reader.read(8192), \
                    new_key, first)
            if data == b'':
                if writer is not None:
                    writer.close()
                return
            writer.write(data)
            await writer.drain()
    except:
        try:
            if writer is not None:
                writer.close()
            return
        except:
            return

def close_conns(conns: list):
    for writer in conns:
        writer.transport.close()

async def main():
    global server

    server = await asyncio.start_server(process_stream,
        host=conf['local_host'], port=conf['local_port'])
    ''' TODO: use logging module in the future '''
    print('{0}:{1} is serving'.format(
        conf['local_host'], conf['local_port']
    ))

    ''' no event-loop exception warning.
        it will be optional in the future '''

    loop = asyncio.get_running_loop()
    loop.set_exception_handler(nop)
    no_check(loop)

    async with server:
            await server.serve_forever()

def init():
    if not check_python_version(3, 7, 0):
        raise ValueError('python version not support')

    global new_key, ctx, rmt_conns, conf

    conf = flags.parse()
    if conf is None:
            conf = config.get_config(CLIENT_SIDE)
    new_key = hashed_key(conf['password'].encode())
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.check_hostname = False
    rmt_conns = []

def entry():
    init()

    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print('exit')

def no_check(loop: asyncio.AbstractEventLoop):
    '''XXX not safe!
    we override a private method of event-loop to avoid raising exception
    after keyboardInterrupte.
    '''
    loop._check_closed = nop

if __name__ == '__main__':
    entry()


#!/usr/bin/env python3
''' Author: github.com/loniceraLeo 
'''

import asyncio
import ssl

from TlsProxy import config, flags
from TlsProxy.config import SERVER_SIDE
from TlsProxy.utils import *

def nop(*args, **kwargs):
    pass

async def process(rd: asyncio.StreamReader, wt: asyncio.StreamWriter):
    ''' we shouldn't use ssl mode because the http proxy
    transports the encrypted data in the tunnel. 
    '''
    data = await rd.read(256)
    raw_adr, _ = mask(data, new_key)
    adr = extract_address(raw_adr)
    host, port = adr.split(':')[0], \
                int(adr.split(':')[1], 10)
    if not is_valid_address(host, port):
        return
    try:
        rmt_rd, rmt_wt = await asyncio.open_connection(
            host=host, port=port) 
    except:
        wt.close()
        return 

    await asyncio.gather(stream_copy(rd, rmt_wt), 
                        stream_copy(rmt_rd, wt))        

async def stream_copy(reader: asyncio.StreamReader, 
    writer: asyncio.StreamWriter, toggle=False):
    first = 0
    try:
        while True:
            data, first = mask(await reader.read(8192), \
                        new_key, first)
            if data == b'':
                if writer:
                    writer.close()
                return
            writer.write(data)
            await writer.drain()
    except:
        if writer:
            writer.close()
        return

def extract_address(raw_data: bytes) -> str:
    length = raw_data[0]
    raw_address = raw_data[1:1+length]
    address = str(bytes(raw_address))[2:-1]
    return address

async def main():
    server = await asyncio.start_server(process, 
    host=conf['server'], port=conf['port'],
    ssl=ctx, ssl_handshake_timeout=30)

    print('{0}:{1} is serving'.format(
        conf['server'], conf['port']
    ))
    '''XXX no event-loop exception warning.
    it will be optional in the future '''
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(nop)
    no_check(loop)

    async with server:
        await server.serve_forever()

def init():
    if not check_python_version(3, 7, 0):
        raise ValueError('python version not support')

    global new_key, ctx, conf
    conf = flags.parse()
    if conf is None:
            conf = config.get_config(SERVER_SIDE)
    new_key = hashed_key(conf['password'].encode())
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ctx.load_cert_chain(conf['certificate'], conf['private-key'])

def entry():
    init()

    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print('exit')

def no_check(loop: asyncio.AbstractEventLoop):
    '''XXX not safe
    '''
    loop._check_closed = nop

if __name__ == '__main__':
    nop()



#!/usr/bin/env python3
''' Author: github.com/loniceraLeo
'''

import os
import json
import ssl
from os import system as sys_call

SERVER_SIDE = 0x0
CLIENT_SIDE = 0x1

def nop(*args, **kwargs):
    pass

def get_config(*args) -> dict:
    files = os.listdir()
    for name in files:
        if name == 'config.json':
            config = read_config(name)
            if not 'private-key' in config and args[0] == SERVER_SIDE:
                raise ValueError('not server config')
            return config
    return create_config(args[0])

def set_config():
    ''' reserve this method for future implementation
    '''
    pass

def read_config(filename: str) -> dict:
    with open('./'+filename, 'r') as f:
        marshalled_data = f.read()
        config = json.loads(marshalled_data)
    return config

def create_config(side: int) -> dict:
    ''' creates a new config.json file and save it in
        current dictionary 
    '''
    config = {}
    config['server']          = input('server: ')
    config['port']            = input('port: ')
    config['password']        = input('password: ')
    if side == CLIENT_SIDE:
        config['local_host']  = '127.0.0.1'
        config['local_port']  = input('local_port: ')
    if side == SERVER_SIDE:
        config['certificate'] = input('certificate: ')
        config['private-key'] = input('private-key: ')
        generate_key_cert(config['private-key'], config['certificate'])
    config['hash-algorithm']  = 'blake2b'
    config['padding-length']  = 256

    jsoned_config = json.dumps(obj=config, indent=4)

    with open('./config.json', 'w') as f:
        f.write(jsoned_config)

    return config

def generate_key_cert(key_file, cert_file, *, format: str=None) -> None:
    ''' if current os support, use openssl.
        if argument format is not specified, create a default
        private key and use it to generate a self-signed certificate.
        it is recommended to use default format 
    '''
    files = os.listdir()
    if key_file in files or cert_file in files:
        return 
    try:
        o_v = ssl.OPENSSL_VERSION_INFO
    except:
        raise ValueError('openssl not support')
    default_key_format = {
        'cipher': 'aes256',
        'length': 2048,
        'no-pharse': True   # generate a private without pass pharse
    }
    default_cert_format = {
        'last-time': 3650  # maximum
    }
    if format:  
        ''' TODO: implement user-specified private key and certificate '''
        pass
    else:
        sys_call('openssl genrsa -{0} -out {1} {2}'.format(\
            default_key_format['cipher'], key_file, \
            default_key_format['length']))
        # no pass pharse:
        sys_call('openssl rsa -in {0} -out {0}'.format(\
            key_file))
        sys_call('openssl req -new -x509 -days {0} -key {1} -out {2}'.\
            format(default_cert_format['last-time'],
            key_file, cert_file))

def search_recursively(dir: str, filename: str) -> str:
    files = os.listdir(dir)
    for file in files:
        if file == filename:
            with open(dir+'/'+file) as f:
                data = json.loads(f.read())
                os.chdir(dir)
                return data
        if is_dir(file):
            return search_recursively(dir+'/'+file, filename)
    return None

def is_dir(filename: str) -> bool:
    if not '.' in filename:
        return True
    return False

if __name__ == '__main__':
    #generate_key_cert('abc.key', 'abc.crt')
    nop()


#!/usr/bin/env python3
''' Author: github.com/loniceraLeo
'''

'''
    TODO: random padding length
    TODO: hash algorithms switch
'''

import random
import sys
from hashlib import blake2b, blake2s

def nop(*args, **kwargs):
    pass

def check_python_version(major, minor, micro) -> bool:
    if sys.version_info < (major, minor, micro):
        return False
    return True

def mask(payload: bytes, pwd: bytes, first=0) -> tuple:
    ''' do mask in xor operation.
        the function ensures an ordered mask-unmask behavior. '''
    data = bytearray(payload)
    for i in range(len(data)):
        data[i] = data[i] ^ pwd[(i+first)%len(pwd)]
    last = (len(data)+first) % len(pwd)
    return (data, last)

def hashed_key(key: bytes, size=32) -> bytes:
    ''' check https://www.blake2.net/ for more information '''
    return blake2b(key, digest_size=size).digest()

def is_valid_address(hostname: str, port: int) -> bool:
    if port != 80 and port != 443:
        return False
    try:
        for i in range(len(hostname)):
            ascii_num = ord(hostname[i])
            if hostname[i] == '.' or hostname[i] == '-' or \
                ascii_num in list(range(48, 58)) or \
                    ascii_num in list(range(65, 91)) or \
                        ascii_num in list(range(97, 123)):
                continue
            else:
                return False
    except:
        return False
    return True

def padding(addr: bytes, padding_length: int) -> bytes:
    addr_length = len(addr)
    if addr_length >= padding_length:
        raise ValueError('padding length over limitation')
    payload = bytearray(b'')
    payload.append(addr_length)
    payload.extend(addr)
    pads = random.randbytes(padding_length-addr_length-1)
    payload.extend(pads)
    return payload

def is_valid_address_test(host, port):
    print(is_valid_address(host, port))

if __name__ == '__main__':
    nop()


#!/usr/bin/env python3
''' Author: github.com/loniceraLeo
'''

import os
import sys

from TlsProxy import config

def nop(*args, **kwargs):
    pass

version = '0.8.0'

dic = {
    '-c': 'c', 
    '-g': 'g',
    '-r': 'r',
    '-h': 'h',
    '-v': 'v'
}

def parse() -> any:
    argv = sys.argv
    for index, item in enumerate(argv):
        try:
            if dic[item] == 'c':
                try:
                    conf = config.read_config(argv[index+1])
                except:
                    print('file not found')
                    exit()
                return conf
            if dic[item] == 'g':
                try:
                    config.generate_key_cert(argv[index+1], 
                        argv[index+2])
                except:
                    return None
                exit()
            if dic[item] == 'r':
                try:
                    conf = config.search_recursively(os.getcwd(),
                            argv[index+1])
                except:
                    print('file not found')
                    exit()
                return conf
            if dic[item] == 'h':
                help_inf()
                exit()
            if dic[item] == 'v':
                print('TlsProxy {0}'.format(version))
                exit()

        except KeyError:
            continue
    return None

def help_inf():
    print('''
Usage of TlsProxy:
    tpclient/tpserver [command] [filename] ...
        
    flag:
        -c filename : use filename as config file
        -g filename1 filename2 : generate a key and a self-signed 
        certificate. Respectively named filename1 and filename2
        -r filename : search filename in current directory recursively
        -h : help information
        -v : version info
    ''')

if __name__ == '__main__':
    help_inf()


from setuptools import setup

with open('./README.md', 'r', encoding='utf-8') as f:
    ld = f.read()

setup(
    name = 'TlsProxys',
    version = '1.0.0',
    description = 'a safe http proxy over tls',
    long_description_content_type = 'text/markdown',
    long_description = ld,
    author = 'github.com/loniceraLeo',
    url = 'https://www.github.com/loniceraLeo/TlsProxy',
    packages = ['TlsProxy'],
    python_requires = '>=3.6.0',
    license = 'GPL',
    entry_points = {
        'console_scripts':
            [
                'tpserver = TlsProxy.server:entry',
                'tpclient = TlsProxy.client:entry'
            ]
    }
)